"""
Writing evet logs into base
"""
import tornado.log
from tornado.queues import Queue
from tornado.options import options


class TorskelEventLogController:
    """
    Class for write events
    """

    def __init__(self):
        self.logger = tornado.log.gen_log
        self.queue = Queue()

    def add_log_event(self, event):
        """
        Put event into queue
        :param event:
        :return:
        """
        if isinstance(event, dict):
            self.logger.debug(event)
            self.queue.put(event)

    # pylint: disable=C0103
    async def write_log_from_queue(self, db, collection_name,
                                   events_writer_func) -> type(None):
        """
        Retrieves events from the queue.
        and performs the insert into the database
        """

        qsize = self.queue.qsize()
        if options.show_log_event_writer:
            self.logger.info('Writing events... queue size = %s', qsize)
        if qsize > 0:
            step = qsize if qsize <= options.task_list_size else \
                options.task_list_size
            inserts_list = [await self.queue.get() for _ in range(step)]
            if inserts_list:
                await events_writer_func(db, collection_name, inserts_list)



"""
Class for getting service status
"""

from torskel.torskel_handler import TorskelHandler


# pylint: disable=W0223
class TorskelPingHandler(TorskelHandler):
    """
    Return status of service
    """

    async def get_ok_status(self):
        """
        Return OK
        :return:
        """
        self.write(self.get_result_dict(code=0, message='Service is up'))
        self.finish()

    async def get(self):
        """
        HTTP GET request
        :return:
        """
        await self.get_ok_status()

    async def post(self):
        """
        http POST request
        :return:
        """
        await self.get_ok_status()



"""
Module contains startup methods
"""
# pylint: disable=C0103
import logging
import platform
import importlib
import tornado
import tornado.log
try:
    from tornado.netutil import bind_unix_socket
    no_unix_socket = False
except ImportError:
    no_unix_socket = True
from tornado.options import options
from tornado.httpserver import HTTPServer

import torskel


# pylint: disable=C0103
logger = tornado.log.gen_log
# hotfix for windows


def _configure_graylog():
    """
    Configuring graylog's handlers
    :return:
    """
    logger.info('Configuring graylog host=%s port=%s', options.graylog_host,
                options.graylog_port)
    try:
        graypy = importlib.import_module('graypy')
        if graypy.__version__[0] < 1:
            handler = graypy.GELFHandler(
                options.graylog_host,
                options.graylog_port,
                localname=options.srv_name
            )
        else:
            handler = graypy.GELFUDPHandler(
                options.graylog_host,
                options.graylog_port,
                localname=options.srv_name
            )
        logging.getLogger("tornado.access").addHandler(handler)
        logging.getLogger("tornado.application").addHandler(handler)
        logging.getLogger("tornado.general").addHandler(handler)
    except ImportError:
        raise ModuleNotFoundError('Required package graypy is missing')


def server_init(server):
    """
    Initializing an application on a port or socket depending on the settings
    :param server:
    :return: None
    """
    logger.info('')
    logger.info('Starting %s v%s', server.server_name, server.server_version)
    logger.info('')
    logger.info('======== Environment info ======== ')
    logger.info('   Python v%s', platform.python_version())
    logger.info('   Tornado v%s', tornado.version)
    logger.info('   Torskel v%s', torskel.version)
    logger.info('   %s %s ', platform.system(), platform.release())
    logger.info('================================== ')

    if options.use_graylog:
        _configure_graylog()

    if options.run_on_socket and not no_unix_socket:

        unix_socket = bind_unix_socket(options.socket_path, 0o666)
        http_server = HTTPServer(server)
        http_server.add_socket(unix_socket)
        logger.info('Running on socket %s', options.socket_path)
    else:

        logger.info('Running on port %s', options.port)
        server.listen(options.port)



"""
Module with JWT functionality
"""

import jwt
import tornado.log

JWT_DEFAULT_OPTIONS = {
    'verify_signature': True,
    'verify_exp': True,
    'verify_nbf': False,
    'verify_iat': True,
    'verify_aud': False
}
DEFAULT_ALGORITM = 'HS256'
INVALID_AUTH_HEADER = "Invalid header authorization"
RESULT_KEY = 'result'
MESSAGE_KEY = 'message'
TOKEN_KEY = 'token'
TOKEN_INFO_KEY = 'token_info'
AUTH_RES_FAIL = {RESULT_KEY: False, MESSAGE_KEY: INVALID_AUTH_HEADER}
AUTH_RES_GOOD = {RESULT_KEY: True, MESSAGE_KEY: ''}
INVALID_TOKEN = {RESULT_KEY: False, MESSAGE_KEY: 'Invalid Token'}
logger = tornado.log.gen_log  # pylint: disable=C0103


# pylint: disable=W0703
def jwt_decode(token: str, secret_key: str, options: dict = None) -> dict:
    """
    Decode JWT token
    :param token:
    :param secret_key:
    :param options:
    :return:
    """
    res = {RESULT_KEY: True, MESSAGE_KEY: ''}
    if options is None:
        options = JWT_DEFAULT_OPTIONS
        try:

            res.update(jwt.decode(token, secret_key, options=options))
        except Exception:
            logger.exception('jwt_decode failed')
            res[RESULT_KEY] = False

    else:
        res[RESULT_KEY] = False
    return res


def jwt_encode(secret_key, payload: dict = None,
               algoritm: str = DEFAULT_ALGORITM):
    """
    encode JWT
    :param secret_key:
    :param payload:
    :param algoritm:
    :return:
    """
    res = jwt.encode(payload=payload, key=secret_key, algorithm=algoritm,
                     ).decode("utf-8")
    return res


def validate_token(auth_header):
    """
    Validate access token
    :param auth_header:
    :return:
    """
    res = AUTH_RES_GOOD
    if auth_header:
        try:
            parts = auth_header.split()
            if parts[0].lower() != 'bearer':
                res = AUTH_RES_FAIL

            elif len(parts) == 1:
                res = AUTH_RES_FAIL
            elif len(parts) > 2:
                res = AUTH_RES_FAIL
            else:
                res[TOKEN_KEY] = parts[1]
        except IndexError:
            logger.exception('_validate_token failed')
            res = INVALID_TOKEN
        except KeyError:
            logger.exception('_validate_token failed')
            res = INVALID_TOKEN
    else:
        res = AUTH_RES_FAIL
    return res


# pylint: disable=W0212
# pylint: disable=W0703
def jwtauth(handler_class):
    """
        Class decorator to check for authorization
    """

    # pylint: disable=W0613

    def wrap_execute(handler_execute):
        def require_auth(handler, kwargs):

            auth = handler.request.headers.get('Authorization')
            token_validation_res = validate_token(auth)

            if token_validation_res.get(RESULT_KEY, False):

                token = token_validation_res.get(TOKEN_KEY)
                try:

                    res = jwt_decode(
                        token, handler.application.get_secret_key()
                    )

                    if not res.get(RESULT_KEY):
                        handler._transforms = []
                        handler.set_status(401)
                        handler.write("Missing authorization")
                        handler.finish()

                except Exception:

                    handler._transforms = []
                    handler.set_status(401)
                    handler.write("Missing authorization")
                    handler.finish()

            else:
                handler._transforms = []
                handler.set_status(401)
                handler.write("Missing authorization")
                handler.finish()

            return True

        def _execute(self, transforms, *args, **kwargs):

            try:
                require_auth(self, kwargs)
            except Exception:
                return False

            return handler_execute(self, transforms, *args, **kwargs)

        return _execute

    handler_class._execute = wrap_execute(handler_class._execute)
    return handler_class



"""
Module with useful string functions
"""
import hashlib
import re
import ipaddress
import datetime
import json
import xmltodict

ALL_HASH_RE_TMPL = r"^(?:[a-fA-F\d]{32,40})$|^(?:[a-fA-F\d]{52,60})$|" \
                   r"^(?:[a-fA-F\d]{92,100})$"

# pylint: disable=C0103
hash_sha224_tmpl = re.compile(r"\b([a-f\d]{56}|[A-F\d]{56})\b")
all_hash_tmpl = re.compile(ALL_HASH_RE_TMPL)
mac_address = re.compile('^' + r'[\:\-]'.join(['([0-9a-f]{2})'] * 6) + '$')


def default_json_dt(json_object):
    """
    Formatting date in JSON
    :param json_object:
    :return:
    """
    if isinstance(json_object, (datetime.date, datetime.datetime)):

        res = json_object.isoformat()
    else:
        res = None
    return res


# pylint: disable=C0103
def is_valid_ip(ip):
    """
    Validate ip address
    :param ip: ip address
    :return: boolean
    """
    res = False
    if isinstance(ip, str):
        try:
            ipaddress.ip_address(ip)
            res = True
        except ValueError:
            res = False
    return res


# pylint: disable=C0103
def is_valid_mac(mac):
    """
    Validate mac address
    :param mac:
    :return: boolean
    """
    res = False
    if isinstance(mac, str):
        if mac:
            res = mac_address.match(mac.lower()) is not None

    return res


def chr_set_null(chr_value):
    """sql util function"""
    return "null" if chr_value is None or not chr_value else \
        ''.join(("'", chr_value, "'"))


def int_set_null(int_value):
    """sql util function"""
    return "null" if int_value is None else str(int_value)


def valid_conversion(val, type_to_convert):
    """
    Checking whether it is possible to convert val to the specified type
    :param val: value
    :param type_to_convert: type
    :return: boolean
    """

    if isinstance(type_to_convert, type):
        try:
            type_to_convert(val)
            res = True
        except ValueError:
            res = False

    else:
        raise TypeError
    return res


def get_hash_str(value, alg='sha224'):
    """
    Return hash from string
    :param value: string
    :param alg: algorithm, default sha224
    :return: hash
    """
    res = None

    try:
        if isinstance(value, str):
            res = getattr(hashlib, alg)(value.encode('utf-8')).hexdigest()
    except UnicodeEncodeError:
        pass

    return res


def is_hash_str(value):
    """
    Checks whether the string is a hash of any algorithm
    :param value:
    :return: boolean
    """
    res = False
    if isinstance(value, str):
        if value:
            res = all_hash_tmpl.match(value) is not None

    return res


def is_number(value):
    """
    Checks whether the value is a number
    :param value:
    :return: bool
    """
    try:
        complex(value)
    except ValueError:
        return False

    return True


def xml_str_to_dict(xml_str: str) -> dict:
    """
    create dict from xml
    :param xml_str:
    :return:
    """
    parsed_xml = xmltodict.parse(xml_str)
    return json.loads(json.dumps(parsed_xml))



"""
Module contains class for work with Redis
"""
import json
import asyncio
import importlib

import tornado.web
import tornado.httpclient
import tornado.log

from tornado.options import options


class RedisApplicationMixin():
    """
    Redis controller
    """
    def __init__(self):
        self.logger = tornado.log.gen_log
        self.redis_connection_pool = None

    @property
    def redis_addr(self) -> str:
        """
        Returns address for connecting to redis
        :return: str
        """
        redis_addr = options.redis_socket if options.use_redis_socket \
            else (options.redis_host, options.redis_port)
        return redis_addr

    @property
    def redis_min_con(self) -> int:
        """
        Returns the minimal size of redis connections pool
        :return: int
        """
        return options.redis_min_con

    @property
    def redis_max_con(self) -> int:
        """
        Returns the maximal size of redis connections pool
        :return: int
        """
        return options.redis_max_con

    @property
    def redis_psw(self) -> str:
        """
        Returns password for connecting to redis
        :return: str
        """
        psw = options.redis_psw

        if psw == "":
            psw = None
        return psw

    @property
    def redis_db(self) -> int:
        """
        Returns number of default redis database
        :return: int
        """
        db_number = options.redis_db
        if db_number == -1:
            db_number = None
        return db_number

    @staticmethod
    def _get_json_util() -> object:
        """
        Returns json_util from bson if exists
        :return:
        """
        try:
            bson = importlib.import_module('bson')

            json_util = bson.json_util
        except ImportError:
            json_util = False
        return json_util

    def init_redis_pool(self):
        """
        Init redis connection pool
        """
        try:
            aioredis = importlib.import_module('aioredis')
            self.redis_connection_pool \
                = asyncio.get_event_loop().run_until_complete(
                    aioredis.create_pool(
                        self.redis_addr,
                        password=self.redis_psw,
                        db=self.redis_db,
                        minsize=self.redis_min_con,
                        maxsize=self.redis_max_con
                    )
                )
        except ImportError:
            raise ImportError('Required package aioredis is missing')

    async def set_redis_exp_val(self, key, val, exp=None, **kwargs):
        """
        Write value to redis
        :param key: key
        :param val: value
        :param exp: Expire time in seconds
        param convert_to_json: bool
        param use_json_utils: bool use json utils from bson

        """

        convert_to_json = kwargs.get('convert_to_json', False)
        use_json_utils = kwargs.get('use_json_utils', False)
        if convert_to_json:
            if use_json_utils:
                json_util = self._get_json_util()
                if json_util:
                    val = json.dumps(val, default=json_util.default)
                else:
                    raise ImportError('Can not import json_util. '
                                      'Module bson is missing')
            else:
                val = json.dumps(val)

        await self.redis_connection_pool.execute('set', key, val)
        if isinstance(exp, int):
            await self.redis_connection_pool.execute('expire', key, exp)

    async def del_redis_val(self, key):
        """
        delete value from redis by key
        :param key: key

        """

        await self.redis_connection_pool.execute('del', key)

    async def get_redis_val(self, key: str, **kwargs):
        """
        get value from redis by key
        :param key: key
        param from_json: loads from json
        param use_json_utils: bool use json utils from bson
        :return: value
        """

        from_json = kwargs.get('from_json', False)
        use_json_utils = kwargs.get('use_json_utils', False)

        val = await self.redis_connection_pool.execute('get', key)
        redis_val = val.decode('utf-8') if val is not None else val
        if redis_val:
            if use_json_utils:
                json_util = self._get_json_util()
                if json_util:
                    res = json.loads(
                        redis_val, object_hook=json_util.object_hook
                    ) if from_json else redis_val
                else:
                    raise ImportError('Can not import json_util. '
                                      'Module bson is missing')

            else:
                res = json.loads(redis_val) if from_json else redis_val
        else:
            res = None
        return res
